import sys,os
import multiprocessing
import time

####################################################
# abstact layer of game

class Polyhedron:
    """
    This is base class of polyhedron. 
    """
    def __init__(self,faces,face_as_edges,edge_as_vertices,layout_hint,layout_hint_type):
        """
        Construct an object from 
        a list of the faces, 
        a dictionary which maps each face to edges contained by it, and
        a dictionary which maps each edge to vertices contained by it.
        """
        self.faces=frozenset(faces)
        self.face_as_edges=dict(face_as_edges)
        self.edge_as_vertices=dict(edge_as_vertices)
        self.layout_hint=[[ row[:] for row in chart] for chart in layout_hint]
        self.layout_hint_type=layout_hint_type

    def get_copied_polyhedron(self):
        p=Polyhedron(self.get_faces(),self.get_face_as_edges(),self.get_edge_as_vertices(),self.get_layout_hint(),self.get_layout_hint_type())

        if 'face_as_vertices' in dir(self):
            p.face_as_vertices=self.get_face_as_vertices()
        if 'vertices' in dir(self):
            p.vertices=self.get_vertices()
        if 'edges' in dir(self):
            p.edges=self.get_edges()
        if 'dict_faces_from_edge' in dir(self):
            p.dict_faces_from_edge=dict(self.dict_faces_from_edge)
        if 'dict_faces_from_vertex' in dir(self):
            p.dict_faces_from_vertex=dict(self.dict_faces_from_vertex)
        if 'dict_edges_from_vertex' in dir(self):
            p.dict_edges_from_vertex=dict(self.dict_faces_from_vertex)

        return p

    def get_euler_characteristic_of(self,faces):
        """
        returns the Euler characteristic for the polyhedron restricted to faces.
        """
        all_faces = self.get_faces()
        set_faces = set([])
        set_edges = set([])
        set_vertices = set([])
        for f in faces:
            if f in all_faces:
                set_faces.add(f)
                set_vertices.update(self.get_vertices_contained_by_a_face(f))
                set_edges.update(self.get_edges_contained_by_a_face(f))
        return len(set_vertices) - len(set_edges) + len(set_faces)

    def get_faces(self):
        """
        return the set of faces.
        """
        return frozenset(self.faces)

    def get_edge_as_vertices(self):
        return dict(self.edge_as_vertices)

    def get_face_as_vertices(self):
        if 'face_as_vertices' in dir(self):
            return dict(self.face_as_vertices)

        f2e=self.get_face_as_edges()
        e2v=self.get_edge_as_vertices()
        f2v={}
        for f in self.get_faces():
            f2v[f]=frozenset([vij for ei in f2e[f] for vij in e2v[ei]])
            
        self.face_as_vertices=f2v
        return dict(f2v)

    def get_face_as_edges(self):
        return dict(self.face_as_edges)

    def get_vertices(self):
        """
        return the set of vertices.
        """
        if 'vertices' in dir(self):
            return frozenset(self.vertices)
        vertices=set([])
        for vs in self.edge_as_vertices.values():
            vertices=vertices.union(vs)
        self.vertices=frozenset(vertices)
        return frozenset(self.vertices)


    def get_edges(self):
        """
        return the set of edges.
        """
        if 'edges' in dir(self):
            return frozenset(self.edges)
        edges=set([])
        for es in  self.face_as_edges.values():
            edges=edges.union(es)
        self.edges=frozenset(edges)
        return frozenset(self.edges)

    def get_vertices_contained_by_an_edge(self,edge):
        """
        return the set of vertices in face.
        """
        dic=self.get_edge_as_vertices()
        return dic[edge]

    def get_edges_containing_a_vertex(self,vertex):
        """
        return the set of vertices in face.
        """
        if 'dict_edges_from_vertex' in dir(self):
            if vertex in self.dict_edges_from_vertex:
                return frozenset(self.dict_edges_from_vertex[vertex])
        self.dict_edges_from_vertex={}
            
        r=set([])
        for e in self.get_edges():
            vs=self.get_vertices_contained_by_an_edge(e)
            if vertex in vs:
                r.add(e)
        self.dict_edges_from_vertex[vertex]=frozenset(r)

        return r

    def get_vertices_contained_by_a_face(self,face):
        """
        return the set of vertices in face.
        """
        dic=self.get_face_as_vertices()
        return dic[face]

    def get_faces_containing_a_vertex(self,vertex):
        """
        return the set of faces contaning vertex.
        """
        if 'dict_faces_from_vertex' in dir(self):
            if vertex in self.dict_faces_from_vertex:
                return frozenset(self.dict_faces_from_vertex[vertex])
        self.dict_faces_from_vertex={}


        r=set([])
        for f in self.get_faces(): 
            vs = self.get_vertices_contained_by_a_face(f)
            if vertex in vs:
                r.add(f)
        self.dict_faces_from_vertex[vertex]=frozenset(r)
                
        return r

    def get_edges_contained_by_a_face(self,face):
        """
        return the set of vertices in face.
        """
        dic=self.get_face_as_edges()
        return dic[face]

    def get_faces_containing_an_edge(self,edge):
        """
        return the set of faces contaning edge.
        """
        if 'dict_faces_from_edge' in dir(self):
            if edge in self.dict_faces_from_edge:
                return self.dict_faces_from_edge[edge]
        else:
            self.dict_faces_from_edge={}

        r=set([])
        for f in self.get_faces(): 
            es = self.get_edges_contained_by_a_face(f)
            if edge in es:
                r.add(f)
        self.dict_faces_from_edge[edge]=frozenset(r)

        return r

    def get_layout_hint(self):
        """
        return layout hint.
        """
        return self.layout_hint

    def get_layout_hint_type(self):
        """
        return layout hint.
        """
        return self.layout_hint_type

class Board:
    def __init__(self,polyhedron,cells_eachplayer=None,n_turn=None):
        if cells_eachplayer==None:
            cells_eachplayer=[[],[]]
        if n_turn==None:
            n_turn=0
        self.polyhedron=polyhedron.get_copied_polyhedron()
        self.cells=[ cells_i[:] for cells_i in cells_eachplayer ]
        self.n_turn=n_turn

    def get_copied_board(self):
        return Board(self.get_polyhedron(),self.cells,self.n_turn)

    def get_polyhedron(self):
        return self.polyhedron.get_copied_polyhedron()

    def get_point(self,player):
        return self.get_polyhedron().get_euler_characteristic_of(self.get_cells_of(player))

    def get_all_cells(self):
        return self.get_polyhedron().get_faces()

    def get_face_cells(self):
        fs=self.get_polyhedron().get_faces()
        r=set([])
        for c in self.get_all_cells():
            if c in fs:
                r.add(c)
        return r

    def get_edge_cells(self):
        es=self.get_polyhedron().get_edges()
        r=set([])
        for c in self.get_all_cells():
            if c in es:
                r.add(c)
        return r

    def get_vertex_cells(self):
        vs=self.get_polyhedron().get_vertices()
        r=set([])
        for c in self.get_all_cells():
            if c in vs:
                r.add(c)
        return r

    def get_empty_cells(self):
        cells=self.get_all_cells()
        return [ cell for cell in cells if self.get_owner_of(cell) < 0 ]

    def n_empty_cells(self):
        n=0
        cells=self.get_all_cells()
        for cell in cells:
            if self.get_owner_of(cell) < 0:
                n += 1
        return n

    def get_an_empty_cell_at(self,n):
        nn=n
        cells=self.get_all_cells()
        for cell in cells:
            if self.get_owner_of(cell) < 0:
                nn -= 1
                if nn < 0:
                    return cell
        return self.get_an_empty_cell_at(n%(n-nn))

    def get_owner_of(self,cell):
        """
        returns player id who put the cell. 
        -1 if no player put the celll.
        """
        for (i,cells_player_i) in enumerate(self.cells):
            if cell in cells_player_i:
                return i
        return -1

    def get_color_of(self,cell):
        """
        returns get_owner_of(cell). 
        (This returns imformation to draw.)
        """
        return self.get_owner_of(cell)

    def is_puttable(self,cell,player):
        if not cell in self.get_empty_cells():
            return False
        return True

    def get_winner(self):
        if self.get_point(0) > self.get_point(1) :
            return 0
        else:
            return 1

    def append_cell(self,player,cell):
        self.cells[player].append(cell)

    def extend_cell(self,player,cells):
        self.cells[player].extend(cells)

    def set_cells(self,player,cells):
        self.cells[player]=cells[:]

    def get_cells_of(self,player):
        return self.cells[player][:]

    def get_face_cells_of(self,player):
        fs=self.get_polyhedron().get_faces()
        r=set([])
        for c in self.get_cells_of(player):
            if c in fs:
                r.add(c)
        return r

    def get_edge_cells_of(self,player):
        es=self.get_polyhedron().get_edges()
        r=set([])
        for c in self.get_cells_of(player):
            if c in es:
                r.add(c)
        return r

    def get_vertex_cells_of(self,player):
        vs=self.get_polyhedron().get_vertices()
        r=set([])
        for c in self.get_cells_of(player):
            if c in vs:
                r.add(c)
        return r

    def pop_cell(self,player):
        return self.cells[player].pop()


class Player:
    def play(self,board,timeout):
        return board.get_an_empty_cell_at(0)

    def played(self,board,cell):
        pass

    def end_game(self,board,winner):
        pass

    def start_game(self,board,player_id):
        self.player_id=player_id
        return True

class Outsider:
    def game_start(self,board):
        pass
    def game_end(self,board,winner):
        pass
    def turn_end(self,board):
        pass

class Referee:
    def __init__(self,board,players,timeouts,outsiders=None):
        self.board=board.get_copied_board()
        self.players=tuple(players)
        self.timeouts=timeouts[:]
        if outsiders:
            self.outsiders=outsiders[:]
        else:
            self.outsiders=[]

    def start_game(self):
        self.announce_game_start()

        while self.board.n_empty_cells() > 0:
            c = self.play_turn()
            self.announce_turn_end(c)

        winner=self.board.get_winner()
        self.announce_game_end(winner)

        return self.get_board()

    def get_board(self):
        return self.board.get_copied_board()

    def announce_game_start(self):
        for i,player in enumerate(self.players):
            board=self.board.get_copied_board()
            player.start_game(board,i)

        for outsider in self.outsiders:
            board=self.board.get_copied_board()
            outsider.game_start(board)

    def announce_turn_end(self,cell):
        for i,player in enumerate(self.players):
            board=self.board.get_copied_board()
            player.played(board,cell)

        for outsider in self.outsiders:
            board=self.board.get_copied_board()
            outsider.turn_end(board,cell)


    def announce_game_end(self,winner):
        for i,player in enumerate(self.players):
            board=self.board.get_copied_board()
            player.end_game(board,winner)
        for outsider in self.outsiders:
            board=self.board.get_copied_board()
            outsider.game_end(board,winner)

    def get_timeout(self,player_id):
        return self.timeouts[player_id]

    def play_turn(self):
        timeout=self.get_timeout(self.board.n_turn % 2)
        player=self.players[self.board.n_turn % 2]
        board=self.board.get_copied_board()
        cell=self.try_calc_in(timeout,lambda:player.play(board,timeout))

        if not self.board.is_puttable(cell,player):
            cell=self.board.get_an_empty_cell_at(0)

        self.board.append_cell(self.board.n_turn % 2,cell)
        self.board.n_turn += 1
        return cell

    def try_calc_in(self,timeout,func):
        if timeout > 0:
            q = multiprocessing.Queue()
            p = multiprocessing.Process(target= (lambda:q.put(func())) )
            p.start()
            p.join(timeout)
            if p.is_alive():
                p.terminate()
                p.join()
            if q.empty():
                return None
            else:
                return q.get()
        else:
            return func()


#################################################
# Rough implementation of polyhedron, player, logger
class SmallPolyhedron(Polyhedron):
    def __init__(self):
        n_row=4
        n_col=4
        self.n_row=n_row
        self.n_col=n_col

        self.faces=[ (i,j) for i in range(n_row) for j in range(n_col) ]
        self.faces.append((n_row,0))

        rep={}
        for i in range(n_row+1):
            for j in range(n_col+1):
                rep[(i,j)] = self.get_equivalent_cell((i,j))

        self.edges=[]
        for i in range(n_row):
            for j in range(n_col):
                e=[rep[(i,j)], rep[(i,j+1)]]
                self.edges.append(frozenset(e))
                e=[rep[(i,j)], rep[(i+1,j)]]
                self.edges.append(frozenset(e))
                e=[rep[(i,j)], rep[(i+1,j+1)]]
                self.edges.append(frozenset(e))
        e=[rep[(n_row-1,0)], rep[(0,n_col-1)],None]
        self.edges.append(frozenset(e))

        self.vertices=[]
        for i in range(n_row):
            for j in range(n_col):
                v=[rep[(i,j)], rep[(i,j+1)], rep[(i+1,j+1)]]
                self.vertices.append(frozenset(v))
                v=[rep[(i,j)], rep[(i+1,j)], rep[(i+1,j+1)]]
                self.vertices.append(frozenset(v))
        v=[rep[(n_row-1,0)], rep[(0,n_col-1)], rep[(n_row,0)], None]
        self.vertices.append(frozenset(v))

        self.face_as_edges={} 
        for f in self.faces:
            es = [ e for e in self.edges if f in e ]
            self.face_as_edges[f]=frozenset(es)

        self.edge_as_vertices={}
        for e in self.edges:
            vs = [ v for v in self.vertices if e.issubset(v)]
            self.edge_as_vertices[e]=frozenset(vs)

        self.edges=frozenset(self.edges)
        self.vertices=frozenset(self.vertices)

        chart=[[self.get_equivalent_cell((i,j)) for j in range(n_col+1)] for i in range(n_row+1)]
        self.layout_hint=[chart]
        self.layout_hint_type="multihex"

    def get_equivalent_cell(self,cell):
        (i,j)=cell
        if i<self.n_row:
            if j<self.n_col:
                return (i,j)
            else:
                return (self.n_row-i,0)
        else:
            if j==0:
                return (i,0)
            else:
                return (0,self.n_col-j)


class NoBrainPlayer(Player):
    def play(self,board,timeout):
        return board.get_an_empty_cell_at(int(time.strftime("%S")))

class SimpleLogger(Outsider):
    def __init__(self,log_prompt=":"):
        self.log_prompt=log_prompt
    def print_log(self,s):
        print(self.log_prompt+s)
    def get_player_string(self,p):
        if p==0:
            return "x"
        else:
            return "o"


    def turn_end(self,board,cell):
        log = self.get_player_string(board.get_owner_of(cell))
        log += ":plays:"
        log += str(cell)
        self.print_log(log)
    def game_end(self,board,winner):
        log = self.get_player_string(winner)
        log += ":wins:"
        self.print_log(log)
        for i in range(2):
            log = self.get_player_string(i)
            log += ":finalscore:"
            log += str(board.get_point(i))
            self.print_log(log)


#################################################
# Plugin loader
class PluginLoader:
    def __init__(self,plugin_dir,relative_path=True):
        if relative_path:
            parent_path = os.path.dirname(os.path.abspath(__file__))
        else:
            parent_path = ""
        plugin_path = os.path.join(parent_path,plugin_dir)
        sys.path.append(plugin_path)
        plugins=[]
        for item in os.listdir(plugin_path):
            if os.path.isdir(os.path.join(plugin_path,item)):
                try:
                    plugins.append(__import__(item))
                except:
                    pass
        self.plugins=plugins

    def get_plugin_list(self):
        return self.plugins

    def get_plugin_name_list(self,key="name"):
        l=[]
        for plugin in self.plugins:
            if 'get_description' in dir(plugin):
                desc=plugin.get_description()
                if key in desc:
                    l.append(desc[key])
                else:
                    l.append("")
            else:
                l.append("")
        return l

#################################################
# Frontend of the game

class ConsoleFrontend:
    def __init__(self):
        self.comment_prompt="## "
        self.players=[HumanPlayerWithConsole(),NoBrainPlayer()]
        self.timeouts=[0,0]
        self.polyhedron=SmallPolyhedron()
        self.new_board=Board
        self.outsiders=[SimpleLogger()]

        self.ai_plugin_loader=PluginLoader('ai_engines')
        self.polyhedron_plugin_loader=PluginLoader('polyhedrons')
        self.board_plugin_loader=PluginLoader('board_factories')
        self.timeout_option=[('no limit',0)]+[(str(i*30)+'S',i*30) for i in range(1,10)]

    def select_player_and_timeout(self):
        players=[]
        timeouts=[]

        for i in range(2):
            message="Select the player "+str(i)

            r=self.select_from_plugin(self.ai_plugin_loader,message,"Select option",label_default="Human")
            if r:
                players.append(r)
                message="Select timelimit for each turn"
                r=self.select_options(message,[l for l,t in self.timeout_option])
                timeouts.append(self.timeout_option[r][1])
            else:
                players.append(HumanPlayerWithConsole())
                timeouts.append(0)
        self.players=players
        self.timeouts=timeouts


    def select_polyhedron(self):
        polyhedron=self.select_from_plugin(self.polyhedron_plugin_loader,"Select a polyhedron","Select option",label_default=None)
        if polyhedron:
            self.polyhedron=polyhedron
        else:
            self.polyhedron=SmallPolyhedron()

    def select_rule(self):
        r=self.select_from_plugin(self.board_plugin_loader,"Select rule","Select option",label_default="Normal rule")
        if r:
            self.new_board=r
        else:
            self.new_board=Board

    def start_game(self):
        board=self.new_board(self.polyhedron)
        referee=Referee(board,self.players,self.timeouts,self.outsiders)
        referee.start_game()

    def print_comment(self,message):
        print(self.comment_prompt+message)

    def select_from_plugin(self,pl,message,message_option,label_default=None):
        plugins=pl.get_plugin_list()
        label=pl.get_plugin_name_list()
        if len(label)==0:
            return None

        if label_default:
            label=[label_default]+label
            r=self.select_options(message,label)-1
        else:
            r=self.select_options(message,label)

        if r<0:
            return None
        p=plugins[r]
        optlist=p.get_option()
        if optlist:
            option_label=[lab for (lab,opt) in optlist]
            r=self.select_options(message_option,option_label)
            opt=optlist[r][1]
        else:
            opt=None
        return p.get_instance(opt)



    def select_options(self,message,label_list):
        for i,label in enumerate(label_list):
            self.print_comment(str(i)+": "+label)
        self.print_comment(" >> "+message+"[Default=0].")

        rc = sys.stdin.readline()
        try:
            i = int(rc)
        except:
            i = 0
        if 0 <= i and i <len(label_list):
            return i
        else:
            return 0


class HumanPlayerWithConsole(Player):
    def __init__(self):
        self.comment_prompt="#"
        self.boardstringformatter = BoardStringFormatter()

    def print_comment(self,s):
        print(self.comment_prompt+s)

    def play(self,board,timeout):
        candidate=list(board.get_empty_cells())
        self.print_comment("--------------------------------------------------------")

        for chart in self.boardstringformatter.get_string(board,candidate):
            for li in chart:
                self.print_comment(" "+li)
        self.print_comment(" Turn of "+self.boardstringformatter.get_player_string(self.player_id))
        self.print_comment(" >> Select the facet [Default=0].")

        rc = sys.stdin.readline()
        try:
            i = int(rc)
        except:
            i = 0
        if 0 <= i and i <len(candidate):
            cell=candidate[i]
        else:
            cell=candidate[0]
        return cell

    
    def start_game(self,board,player_id):
        self.player_id=player_id
        if board.get_polyhedron().get_layout_hint_type() == "multihex":
            return True
        else:
            return False

    def end_game(self,board,winner):
        self.print_comment("--------------------------------------------------------")
        for chart in self.boardstringformatter.get_string(board,[]):
            for li in chart:
                self.print_comment(" "+li)
        if self.player_id == winner:
            self.print_comment(" You, the player "+self.boardstringformatter.get_player_string(self.player_id) +", win.")
        else:
            self.print_comment(" You, the player "+self.boardstringformatter.get_player_string(self.player_id) +", loose.")
        self.print_comment("--------------------------------------------------------")
        


    def get_string_of_board(self,board):
        layout_hint=board.polyhedron.get_layout_hint()
        layout_hint_type=board.polyhedron.get_layout_hint_type()

        r=[]

        if layout_hint_type == "multihex":
            faces=[cij for chart in layout_hint for ci in chart for cij in ci]
            faces.sort()
            face_name_dict=dict([(fi,"") for fi in faces])
            faces=list(set([fi for fi,fj in zip(faces,faces[1:]) if fi == fj]))
            faces.sort()
            for i,fi in enumerate(faces):
                sufix= str(i//26) if i//26 > 0 else ""
                face_name_dict[fi]=chr(ord('A')+i%26)+sufix

            for chart in layout_hint:
                r.append(self.get_board_string_from_chart_hex(chart,board,face_name_dict))
        elif layout_hint_type == "multisquare":
            faces=[cij for chartf,chartv in layout_hint for ci in chartf for cij in ci]
            faces.sort()
            face_name_dict=dict([(fi,"") for fi in faces])
            faces=list(set([fi for fi,fj in zip(faces,faces[1:]) if fi == fj]))
            faces.sort()
            for i,fi in enumerate(faces):
                sufix= str(i//26) if i//26 > 0 else ""
                face_name_dict[fi]=chr(ord('A')+i%26)+sufix

            for chart in layout_hint:
                r.append(self.get_board_string_from_chart_square(chart,board,face_name_dict))
        return r


    def get_board_string_from_chart_square(self,chart,board,face_name_dict):
        chartf,chartv=chart
        l=max([len(ci) for ci in chartf])
        ch=[ ci[:]+(l-len(ci))*[None] for ci in chartf ]

        nullline=[(l+1)*[None]]
        max_len=0

        emptycells=board.get_empty_cells()
        select_key_dict=dict([(fi,str(i)) for i,fi in enumerate(emptycells)])

        max_len=max([len(self.get_face_string(board,cij,face_name_dict,select_key_dict)) for ci in ch for cij in ci])

        facestring=[[ self.get_face_string(board,cij,face_name_dict,select_key_dict,min_size=max_len) for cij in ci] for ci in ch]
        rows=[]

        vcs=board.get_vertex_cells()
        vertex_string_dict=dict([(cell,select_key_dict[cell] if cell in emptycells else self.get_player_string(board.get_color_of(cell))) for cell in vcs])
        max_len_v=max([len(i) for i in vertex_string_dict.values()]+[1])

        barI=[[ True if cij or cik else False for cij,cik in zip([None]+ci,ci+[None])] for ci in ch]

        barH=[[ True if cik or cjk else False for cik,cjk in zip(ci,cj)] for ci,cj in zip(nullline+ch,ch+nullline)]

        xs=[[ bij or bjk for bij,bjk in zip(bi,bj)] for bi,bj in zip(barI+nullline,nullline+barI)]
        xs=[[ " " if not xij else (vertex_string_dict[cij] if cij in vertex_string_dict.keys() else "+") for cij,xij in zip(ci,xi)] for ci,xi in zip(chartv,xs)]


        barIs=[[ "|"+(max_len_v-1)*" " if bij else (max_len_v)*" " for bij in bi] for bi in barI]

        barHs=[[ (max_len_v-len(xij)+ max_len)*"-" if bij else (max_len_v-len(xij)+max_len)*" " for bij,xij in zip(bi,xi)] for bi,xi in zip(barH,xs)]


        hrs=[]
        for i in range(len(ch)+1):
            rowi=""
            for j in range(l):
                rowi+=xs[i][j]
                rowi+=barHs[i][j]
            rowi+=xs[i][-1]
            hrs.append(rowi)

        rows=[]
        for bi,fi in zip(barIs,facestring):
            rowi=""
            for bij,fij in zip(bi,fi+[""]):
                rowi+=bij+fij
            rows.append(rowi)
            
        lines=[]
        lines.append(hrs[0])
        for hi,li in zip(hrs[1:],rows):
            lines.append(li)
            lines.append(hi)

    
        flag=True

        while flag:
            for li in lines:
                if not li.startswith(" "):
                    flag=False
            if flag:
                for i,li in enumerate(lines):
                    lines[i]=li[1:]
        return lines




class BoardStringFormatter:
    def __init__(self):
        self.player_string=("x","o","")
        self.formatable_layout_hint_type=["multihex"]

    def get_formatable_layout_hint_type():
        return self.formatable_layout_hint_type

    def trim_spaces(self,lines):
        """
        remove spaces from list of strings.

        >>> bs=BoardStringFormatter()
        >>> lines=['  abc','   b','  a c']
        >>> bs.trim_spaces(lines)
        >>> lines
        ['abc', ' b', 'a c']
        """
        while True:
            for li in lines:
                if not li.startswith(" "):
                    return 
            for i,li in enumerate(lines):
                lines[i]=li[1:]

    def get_zigzag_serialized_list(self,llist):
        nn=[len(li) for li in llist]
        r=[]
        for j in range(max(nn)):
            for (li,ni) in zip(llist,nn):
                if j<ni:
                    r.append(li[j])
        return r

    def get_player_string(self,p):
        if p==0:
            return self.player_string[0]
        elif p==1:
            return self.player_string[1]
        else:
            return self.player_string[-1]

    def get_string(self,board,list_of_empty_cells):
        layout_hint_type=board.polyhedron.get_layout_hint_type()
        layout_hint=board.polyhedron.get_layout_hint()
        select_key_dict=self.get_select_key_dict(list_of_empty_cells)
        r=[]
        if layout_hint_type == "hex":
            layout_hint=board.polyhedron.get_layout_hint()
            face_name_dict=self.get_equivalent_cell_string_dict(layout_hint)
            r=self.get_string_hex_layout(layout_hint,board,face_name_dict,select_key_dict)
        elif layout_hint_type == "multihex":
            r=self.get_string_multihex_layout(board,select_key_dict)
        elif layout_hint_type == "multixsquare":
            faces=[ fij for chart in layout_hint for fi in chart[0] for fij in fi]
            face_name_dict=self.get_equivalent_cell_string_dict(faces)
            for chart in layout_hint:
                r.append(self.get_string_x_square_layout(chart,board,face_name_dict,select_key_dict))
        elif layout_hint_type == "xsquare":
            faces=[ fij for fi in layout_hint[0] for fij in fi]
            face_name_dict=self.get_equivalent_cell_string_dict(faces)
            r.append(self.get_string_x_square_layout(layout_hint,board,face_name_dict,select_key_dict))
        return r

    def get_equivalent_cell_string_dict(self,cells):
        cell2label={}
        for ci in cells:
            cell2label[ci] = 0
        for ci in cells:
            cell2label[ci] += 1
        cell2label[None]=0
        i = 0
        k = cell2label.keys()
        try:
            k.sort()
        except:
            pass
        for ci in k:
            if cell2label[ci]>1:
                sufix = str(i//26) if i//26 > 0 else ""
                cell2label[ci]=chr(ord('A')+i%26)+sufix
                i += 1
            else:
                cell2label[ci]=""
        return cell2label

    def get_select_key_dict(self,empty_cells):
        r=dict([(fi,str(i)) for i,fi in enumerate(empty_cells)])
        return r
    
    def get_centered_string(self,s,size,left_spacer=" ",right_spacer=" "):
        tl=len(s)
        ll=((size-tl)//2)
        rl=size-ll*len(left_spacer)-tl
        return ll*left_spacer+s+rl*right_spacer

    def get_label(self,board,face,face_name_dict,select_key_dict,min_size=3):
        r=""
        if face:
            if face in select_key_dict:
                r+=select_key_dict[face]
            r+=self.get_player_string(board.get_color_of(face))
            if face in face_name_dict:
                if face_name_dict[face]:
                    r+="("+face_name_dict[face]+")"
        return self.get_centered_string(r,min_size)

    def get_face_string(self,board,face,face_name_dict,select_key_dict,min_size=3):
        return self.get_label(board,face,face_name_dict,select_key_dict,min_size)
    def get_vertex_string(self,board,face,face_name_dict,select_key_dict,min_size=3):
        r=self.get_label(board,face,face_name_dict,select_key_dict,0)
        if r == "":
            r = "+"
        return r+((min_size-len(r))*" ")
    def get_virtical_rule_string(self,board,face,face_name_dict,select_key_dict,min_size=3):
        r=self.get_label(board,face,face_name_dict,select_key_dict,1)
        r = "|"+r
        return r+((min_size-len(r))*" ")

    def get_horizontal_rule_string(self,board,face,face_name_dict,select_key_dict,min_size=3):
        r=self.get_label(board,face,face_name_dict,select_key_dict,1)
        r = "-"+r+"-"
        return r+((min_size-len(r))*"-")


    def get_string_multihex_layout(self,board,select_key_dict):
        layout_hint=board.get_polyhedron().get_layout_hint()
        cells=[cij for chart in layout_hint for ci in chart for cij in ci]
        face_name_dict=self.get_equivalent_cell_string_dict(cells)            
        r=[]
        for chart in layout_hint:
            r.append(self.get_string_hex_layout(chart,board,face_name_dict,select_key_dict))
        return r

    def get_string_hex_layout(self,chart,board,face_name_dict,select_key_dict):
        l=max([len(ci) for ci in chart])
        ch=[ ci[:]+(l-len(ci))*[None] for ci in chart ]

        nullline=[l*[None]]
        max_len=0

        max_len=max([len(self.get_label(board,cij,face_name_dict,select_key_dict)) for ci in ch for cij in ci])
        llen=(max_len-1)//2
        rlen=max_len-1-llen
        facestring=[[ self.get_label(board,cij,face_name_dict,select_key_dict,min_size=max_len) for cij in ci] for ci in ch]
        rows=[]

        barI=[[ True if cij or cik else False for cij,cik in zip([None]+ci,ci+[None])] for ci in ch]
        barJ=[[ True if cik or cjk else False for cik,cjk in zip(ci,cj)] for ci,cj in zip(nullline+ch,ch+nullline)]
        barL=[[ True if cik or cjk else False for cik,cjk in zip(ci+[None],[None]+cj)] for ci,cj in zip(nullline+ch,ch+nullline)]
        xtops=[[ "+" if bij else "-" if cij else " " for bij,cij in zip(bi+[None],ci+[None])] for bi,ci in zip(barI+nullline,nullline+ch)]
        xbots=[[ "+" if bij else "-" if cij else " " for bij,cij in zip(bi+[None],[None]+ci)] for bi,ci in zip(nullline+barI,ch+nullline)]
        barIs=[[ "|" if bij else " " for bij in bi] for bi in barI]
        barLs=[[ llen*"-" if bij else llen*" " for bij in bi] for bi in barL]
        barJs=[[ rlen*"-" if bij else rlen*" " for bij in bi]+[""] for bi in barJ]

        hrs=[]
        for i in range(len(ch)+1):
            rowi=""
            for j in range(l+1):
                rowi+=xbots[i][j]
                rowi+=barLs[i][j]
                rowi+=xtops[i][j]
                rowi+=barJs[i][j]
            hrs.append(rowi)
        rows=[]

        for bi,fi in zip(barI,facestring):
            bars=[ "|" if bij else " " for bij in bi]
            rowi=""
            for bj,fij in zip(bars,fi+[""]):
                rowi+=bj+fij
            rows.append(rowi)
            
        lines=[]
        indent=""
        lines.append(hrs[0])
        indent+=(1+llen)*" "
        for hi,li in zip(hrs[1:],rows):
            lines.append(indent+li)
            lines.append(indent+hi)
            indent+=(1+llen)*" "
        self.trim_spaces(lines)
        return lines

    def get_string_x_square_layout(self,layouthint,board,face_name_dict,select_key_dict):
        (face,hedge,vedge,vertex) = layouthint

        face_label = [[self.get_label(board,fij,face_name_dict,select_key_dict) for fij in fi ]  for fi in face]
        hedge_spacer = [[ "-" if Hij else " " for Hij in Hi] for Hi in hedge]
        hedge_label = [[ "-" if Hij else "" for Hij in Hi] for Hi in hedge]
        vedge_spacer = [[ "|" if Iij else " " for Iij in Ii] for Ii in vedge]
        vedge_label = [[ "|" if Iij else "" for Iij in Ii] for Ii in vedge]
        vertex_label = [[self.get_vertex_string(board,vij,face_name_dict,select_key_dict,min_size=1) for vij in vi] for vi in vertex]

        vertex_and_vedge_max_len = [[ max(len(Iij),len(vij)) for (Iij,vij) in zip(Ii,vi)]for (Ii,vi) in zip(vedge_label,vertex_label)]
        vertex_and_vedge_max_len = [ max([mi[j] for mi in vertex_and_vedge_max_len]) for j in range(len(vertex_and_vedge_max_len[0]))]

        face_and_hedge_max_len = [[max(len(Hij),len(fij)) for (Hij,fij) in zip(Hi,fi)] for (Hi,fi) in zip(hedge_label,face_label)]
        face_and_hedge_max_len = [ max([mi[j] for mi in face_and_hedge_max_len]) for j in range(len(face_and_hedge_max_len))]

        face_string = [[ self.get_centered_string(fij,mj," "," ") for (fij,mj) in zip(fi,face_and_hedge_max_len)] for fi in face_label]
        hedge_string = [[ self.get_centered_string(Hij,mj,sij,sij) for ((Hij,sij),mj) in zip(zip(Hi,si),face_and_hedge_max_len)] for (Hi,si) in zip(hedge_label,hedge_spacer)]

        vertex_string = [[ self.get_centered_string(vij,mj,"",sij) for ((vij,sij),mj) in zip(zip(vi,si+[" "]),vertex_and_vedge_max_len)] for (vi,si) in zip(vertex_label,hedge_spacer)]

        vedge_string = [[ self.get_centered_string(Iij,mj,""," ") for (Iij,mj) in zip(Ii,vertex_and_vedge_max_len)] for Ii in vedge_label]


        face_and_vedge_line=[ "".join(self.get_zigzag_serialized_list([Ii,fi])) for (Ii, fi) in zip(vedge_string,face_string)]

        vertex_and_hedge_line = [ "".join(self.get_zigzag_serialized_list([vi,Hi])) for (Hi, vi) in zip(hedge_string,vertex_string)]

        lines = self.get_zigzag_serialized_list([vertex_and_hedge_line,face_and_vedge_line])

        self.trim_spaces(lines)
        return [lines]


def main_console():
    c=ConsoleFrontend()
    while True:
        c.select_player_and_timeout()
        c.select_polyhedron()
        c.select_rule()
        c.start_game()
        break

def test():
    """
    >>> a=1+2

    >>> a-3
    0
    """
    import doctest
    doctest.testmod()

if __name__ == "__main__":
    if len(sys.argv)>1:
        if sys.argv[1]=="--test":
            test()
        else:
            main_console()
    else:
        main_console()




